<h1>Metadata</h1>
<p><strong>@@@ApplicationName@@@ Downloads</strong>, extracted on <strong>@@@DateExtracted@@@</strong></p>
<p>This file contains @@@StoryCount@@@ stories. They consist of <strong>@@@ChapterCount@@@</strong> chapters and <strong>@@@WordCount@@@</strong> words in total. They have all been extracted on <strong>@@@DateExtracted@@@</strong>, using <a href = "@@@ApplicationURL@@@"><strong>@@@ApplicationName@@@ @@@ApplicationVersion@@@</strong></a>.</p>