<h1>Metadata</h1>
<p><strong>@@@Title@@@</strong> by <strong>@@@Author@@@</strong></p>
<p><em>@@@Summary@@@</em></p>
<p>This story has been published on <strong>@@@DatePublished@@@</strong>. It has been most recently updated on <strong>@@@DateUpdated@@@</strong>. It consists of <strong><span id = "StoryChapterCount">@@@ChapterCount@@@</span></strong> chapter(s) and <strong>@@@WordCount@@@</strong> words. It can be found under the following URL:</p>
<p><a href = "@@@URL@@@"><span id "StoryURL">@@@URL@@@</span></a></p>
<p>This story has been extracted on <strong>@@@DateExtracted@@@</strong>, using <a href = "@@@ApplicationURL@@@"><strong>@@@ApplicationName@@@ @@@ApplicationVersion@@@</strong></a>.</p>